/*
  MPU6050_quaternion.c
  MPU6050 quaternion test serial communication with Processing 1.5.1 
*/
#include "simpletools.h"

#define twoKpDef 0.010f; // (2.0f * 0.5f) // 2 * proportional gain
#define twoKiDef 0.002f; // (2.0f * 0.1f) // 2 * integral gain
#define RAD2DEG 57.2957795130823
#define DEG2RAD 0.01745329251994329576923690768489 

unsigned int devAddr = 0x68;
unsigned char INT_STATUS = 0x3A;

float gyro_off_x, gyro_off_y, gyro_off_z;
float twoKp = twoKpDef; // 2 * proportional gain (Kp)
float twoKi = twoKiDef; // 2 * integral gain (Ki)
float q0 = 1.0f, q1 = 0.0f, q2 = 0.0f, q3 = 0.0f;
float integralFBx = 0.0f,  integralFBy = 0.0f, integralFBz = 0.0f;
float sampleFreq;

char buffer[14];

//float val[6];
float q[4];
//float euler[3];
//float ypr[3];

i2c *MPU6050;

int main(){
  MPU6050 = i2c_newbus(0,  1,   0); 

  initialize();

  zeroGyro();

  while(1){
    //getValues(val);
    //print("ax = %f \n", val[0]);
    //print("ay = %f \n", val[1]);
    //print("az = %f \n", val[2]);
    //print("gx = %f \n", val[3]);
    //print("gy = %f \n", val[4]);
    //print("gz = %f \n", val[5]);

    getQ(q);
    //print("q0 = %f \n", q[0]);
    //print("q1 = %f \n", q[1]);
    //print("q2 = %f \n", q[2]);
    //print("q3 = %f \n", q[3]);
    serialPrintFloatArr(q, 4);    
    print("\n");

    //getEuler(euler);
    //print("psi = %f \n", euler[0]);
    //print("theta = %f \n", euler[1]);
    //print("phi = %f \n", euler[2]);

    //getYawPitchRoll(ypr);
    //print("yaw = %f \n", ypr[0]);
    //print("pitch = %f \n", ypr[1]);
    //print("roll = %f \n", ypr[2]);

  }
}

void initialize(){
  unsigned char PWR_MGMT_1 = 0x6B;
  unsigned char PowerMgmt = 0b00000001;
  unsigned char SMPLRT_DIV = 0x19;
  unsigned char SampleRate = 0b00000001;
  unsigned char CONFIG = 0x1A;
  unsigned char DLP = 0b00000011;
  unsigned char GYRO_CONFIG = 0x1B;
  unsigned char GyroFS = 0b00011000;
  unsigned char ACCEL_CONFIG = 0x1C;
  unsigned char AccelFS = 0b00010000;

 /*
  i2c_out(MPU6050, devAddr, PWR_MGMT_1, 1, &PowerMgmt, 1);
  i2c_out(MPU6050, devAddr, SMPLRT_DIV, 1, &SampleRate, 1);
  i2c_out(MPU6050, devAddr, CONFIG, 1, &DLP, 1);
  i2c_out(MPU6050, devAddr, GYRO_CONFIG, 1, &GyroFS, 1);
  i2c_out(MPU6050, devAddr, ACCEL_CONFIG, 1, &AccelFS, 1);

  while(i2c_busy(MPU6050, devAddr)); 
*/
  int n;
  n = i2c_out(MPU6050, devAddr, PWR_MGMT_1, 1, &PowerMgmt, 1);
  n += i2c_out(MPU6050, devAddr, SMPLRT_DIV, 1, &SampleRate, 1);
  n += i2c_out(MPU6050, devAddr, CONFIG, 1, &DLP, 1);
  n += i2c_out(MPU6050, devAddr, GYRO_CONFIG, 1, &GyroFS, 1);
  n += i2c_out(MPU6050, devAddr, ACCEL_CONFIG, 1, &AccelFS, 1);
 
  if(n != 15){
    print("Bytes should be 15, but was %d,", n);
    while(1);
  } 
}

void zeroGyro(){
  const int totSamples = 25;
  float tmpOffsets[] = {0,0,0};
  for(int i = 0; i < totSamples; i++){
    int ready = 0;
    while(!ready){
      i2c_in(MPU6050, devAddr, INT_STATUS, 1, &ready, 1);
      ready = 1 & ready;
    }
    for(int i = 0; i < 6; i++){ 
      int regAddr = 0x43 + i;
      i2c_in(MPU6050, devAddr, regAddr, 1, &buffer[i], 1);
    }
    tmpOffsets[0] += (buffer[0] << 8) + buffer[1];
    tmpOffsets[1] += (buffer[2] << 8) + buffer[3];
    tmpOffsets[2] += (buffer[4] << 8) + buffer[5];
  }
  gyro_off_x = tmpOffsets[0] / totSamples;
  gyro_off_y = tmpOffsets[1] / totSamples;
  gyro_off_z = tmpOffsets[2] / totSamples;
  //print("gyro_off_x = %f \n", gyro_off_x);
  //print("gyro_off_y = %f \n", gyro_off_y);
  //print("gyro_off_z = %f \n", gyro_off_z);
}

void getValues(float * values){ 
  float accgyroval[6];
  int ready = 0;
  while(!ready){
    i2c_in(MPU6050, devAddr, INT_STATUS, 1, &ready, 1);
    ready = 1 & ready;
  }
  for(int i = 0; i < 14; i++){
    int regAddr = 0x3B + i;
    i2c_in(MPU6050, devAddr, regAddr, 1, &buffer[i], 1);
  }
  accgyroval[0] = (buffer[0] << 8) + buffer[1];
  accgyroval[1] = (buffer[2] << 8) + buffer[3];
  accgyroval[2] = (buffer[4] << 8) + buffer[5];
  accgyroval[3] = (buffer[8] << 8) + buffer[9];
  accgyroval[4] = (buffer[10] << 8) + buffer[11];
  accgyroval[5] = (buffer[12] << 8) + buffer[13];
  accgyroval[3] -= gyro_off_x;
  accgyroval[4] -= gyro_off_y;
  accgyroval[5] -= gyro_off_z;
  for(int i = 0; i<6; i++){
    if(i < 3){
      values[i] = accgyroval[i];
    }else{
      values[i] = accgyroval[i] / 16.4f;
    }
  }
}

void AHRSupdate(float gx, float gy, float gz, float ax, float ay, float az){
  float recipNorm;
  float halfex = 0.0f, halfey = 0.0f, halfez = 0.0f;
  float qa, qb, qc;

  if((ax != 0.0f) && (ay != 0.0f) && (az != 0.0f)){
    float halfvx, halfvy, halfvz;
    
    // Normalise accelerometer measurement
    recipNorm = pow((ax * ax + ay * ay + az * az), -0.5);
    ax *= recipNorm;
    ay *= recipNorm;
    az *= recipNorm;
    
    // Estimated direction of gravity
    halfvx = q1*q3 - q0*q2;
    halfvy = q0*q1 + q2*q3;
    halfvz = q0*q0 - 0.5f + q3*q3;
  
    // Error is sum of cross product between estimated direction and measured direction of field vectors
    halfex += (ay * halfvz - az * halfvy);
    halfey += (az * halfvx - ax * halfvz);
    halfez += (ax * halfvy - ay * halfvx);
  }

  // Apply feedback only when valid data has been gathered from the accelerometer or magnetometer
  if(halfex != 0.0f && halfey != 0.0f && halfez != 0.0f){
    // Compute and apply integral feedback if enabled
    if(twoKi > 0.0f){
      integralFBx += twoKi * halfex * (1.0f / sampleFreq);  // integral error scaled by Ki
      integralFBy += twoKi * halfey * (1.0f / sampleFreq);
      integralFBz += twoKi * halfez * (1.0f / sampleFreq);
      gx += integralFBx;  // apply integral feedback
      gy += integralFBy;
      gz += integralFBz;
    }
    else{
      integralFBx = 0.0f; // prevent integral windup
      integralFBy = 0.0f;
      integralFBz = 0.0f;
    }

    // Apply proportional feedback
    gx += twoKp * halfex;
    gy += twoKp * halfey;
    gz += twoKp * halfez;
  }
  
  // Integrate rate of change of quaternion
  gx *= (0.5f * (1.0f / sampleFreq));   // pre-multiply common factors
  gy *= (0.5f * (1.0f / sampleFreq));
  gz *= (0.5f * (1.0f / sampleFreq));
  qa = q0;
  qb = q1;
  qc = q2;
  q0 += (-qb * gx - qc * gy - q3 * gz);
  q1 += (qa * gx + qc * gz - q3 * gy);
  q2 += (qa * gy - qb * gz + q3 * gx);
  q3 += (qa * gz + qb * gy - qc * gx);
  
  // Normalise quaternion
  recipNorm = pow((q0 * q0 + q1 * q1 + q2 * q2 + q3 * q3), -0.5);
  q0 *= recipNorm;
  q1 *= recipNorm;
  q2 *= recipNorm;
  q3 *= recipNorm;
}

void getQ(float * q){
  unsigned long lastUpdate, now;
  float val[6];
  getValues(val);
  //now = micros();
  //now = CNT / (CLKFREQ / 1000000.0);

  static unsigned long _time = 0;
  unsigned long rc = (CNT - _time);
  _time = CNT;
  now = rc/(CLKFREQ/1000000);
  
  //print("now = %d \n", now);
  //print("lastUpdate = %d \n", lastUpdate); 
  sampleFreq = 1.0 / ((now - lastUpdate) / 1000000.0);
  lastUpdate = now;
  //print("sampleFreq = %f \n", sampleFreq);

  // gyro values are expressed in deg/sec, the * DEG2RAD will convert it to radians/sec
  AHRSupdate(val[3] * DEG2RAD, val[4] * DEG2RAD, val[5] * DEG2RAD, val[0], val[1], val[2]);
  
  q[0] = q0;
  q[1] = q1;
  q[2] = q2;
  q[3] = q3;
}

void getEuler(float * angles){
  float q[4]; // quaternion
  getQ(q);
  angles[0] = atan2(2 * q[1] * q[2] - 2 * q[0] * q[3], 2 * q[0]*q[0] + 2 * q[1] * q[1] - 1) * RAD2DEG; // psi
  angles[1] = -asin(2 * q[1] * q[3] + 2 * q[0] * q[2]) * RAD2DEG; // theta
  angles[2] = atan2(2 * q[2] * q[3] - 2 * q[0] * q[1], 2 * q[0] * q[0] + 2 * q[3] * q[3] - 1) * RAD2DEG; // phi
}


void getYawPitchRoll(float * ypr){
  float q[4]; // quaternion
  float gx, gy, gz; // estimated gravity direction
  getQ(q);
  
  gx = 2 * (q[1]*q[3] - q[0]*q[2]);
  gy = 2 * (q[0]*q[1] + q[2]*q[3]);
  gz = q[0]*q[0] - q[1]*q[1] - q[2]*q[2] + q[3]*q[3];

  ypr[0] = atan2(2 * q[1] * q[2] - 2 * q[0] * q[3], 2 * q[0]*q[0] + 2 * q[1] * q[1] - 1) * RAD2DEG;
  ypr[1] = atan(gx / sqrt(gy*gy + gz*gz)) * RAD2DEG;
  ypr[2] = atan(gy / sqrt(gx*gx + gz*gz)) * RAD2DEG;
}

void serialPrintFloatArr(float * arr, int length){
  for(int i=0; i<length; i++){
    serialFloatPrint(arr[i]);
    print(",");
  }
}

void serialFloatPrint(float f){
  int8_t * b = (int8_t *) &f;
  for(int i=0; i<4; i++){
    int8_t b1 = (b[i] >> 4) & 0x0f;
    int8_t b2 = (b[i] & 0x0f);
    print("%x",b1);
    print("%x",b2);
//    char c1 = (b1 < 10) ? ('0' + b1) : 'A' + b1 - 10;
//    char c2 = (b2 < 10) ? ('0' + b2) : 'A' + b2 - 10;
//    print("%c",c1);
//    print("%c",c2);
  }
}
